from typing import Union
from ._daemon_connection import _daemon_url
from ._misc import _http_post_json


def _set(key: Union[str, dict, list], value: Union[str, dict, list]):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/mutable/set'
    x = _http_post_json(url, dict(
        key=key,
        value=value
    ), headers=headers)
    if not x['success']:
        raise Exception(f'Unable to set value for key: {key}')

def _get(key: Union[str, dict, list]):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/mutable/get'
    x = _http_post_json(url, dict(
        key=key
    ), headers=headers)
    if not x['success']:
        raise Exception(f'Unable to get value for key: {key}')
    found = x['found']
    if found:
        return x['value']
    else:
        return None

def _delete(key: Union[str, dict, list]):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/mutable/delete'
    x = _http_post_json(url, dict(
        key=key
    ), headers=headers)
    if not x['success']:
        raise Exception(f'Unable to delete value for key: {key}')


import os
import time
import tempfile
from typing import List, Union, cast
from ._misc import _http_get_json

def _daemon_port():
    return os.getenv('KACHERY_DAEMON_PORT', 20431)

def _daemon_host():
    return os.getenv('KACHERY_DAEMON_HOST', 'localhost')

_client_auth_code_info = {
    'timestamp': 0,
    'code': ''
}

def _get_client_auth_code():
    elapsed = time.time() - _client_auth_code_info['timestamp']
    if elapsed > 60:
        _client_auth_code_info['code'] = _read_client_auth_code()
        _client_auth_code_info['timestamp'] = time.time()
    return _client_auth_code_info['code']

def _read_client_auth_code():
    ksd = _kachery_storage_dir()
    p = f'{ksd}/client-auth'
    if not os.path.isfile(p):
        raise Exception(f'Unable to find client auth file (perhaps daemon is not running): {p}')
    try:
        with open(p, 'r') as f:
            client_auth_code = f.read()
    except:
        raise Exception(f'Unable to read client auth file. Perhaps you do not have permission to access this daemon.')
    return client_auth_code


def _daemon_url(daemon_port=None, daemon_host=None, no_client_auth=False):
    if daemon_port is not None:
        port = daemon_port
    else:
        port = _daemon_port()
    if daemon_host is not None:
        host = daemon_host
    else:
        host = _daemon_host()
    if not no_client_auth:
        headers = {
            'KACHERY-CLIENT-AUTH-CODE': _get_client_auth_code()
        }
    else:
        headers = {}
    return f'http://{host}:{port}', headers

class _probe_result:
    def __init__(self, x: dict):
        self.probe_response = cast(dict, x)
        self.node_id = cast(str, x['nodeId'])
        ksd = os.getenv('KACHERY_STORAGE_DIR', cast(Union[str, None], x['kacheryStorageDir'] or None))
        if ksd is None:
            raise Exception('No kachery storage directory.')
        if not os.path.exists(ksd):
            raise Exception(f'Kachery storage directory does not exist: {ksd}')
        fname = f'{ksd}/kachery-node-id'
        if not os.path.exists(fname):
            raise Exception(f'File does not exist (perhaps daemon is not running or perhaps you are using an inconsistent version of kachery between daemon and client): {fname}')
        with open(fname, 'r') as f:
            node_id_from_file = f.read()
            if node_id_from_file != self.node_id:
                raise Exception(f'Inconsistent node ID between running daemon and kachery storage directory: {node_id_from_file} <> {self.node_id} ({fname})')
        self.kachery_storage_dir = ksd

class _buffered_probe_data:
    timestamp: float=0
    result: Union[None, _probe_result]=None

def _buffered_probe_daemon(daemon_port=None):
    elapsed_since_last = time.time() - _buffered_probe_data.timestamp
    if elapsed_since_last <= 10:
        return _buffered_probe_data.result
    res = _probe_daemon(daemon_port=daemon_port)
    _buffered_probe_data.timestamp = time.time()
    _buffered_probe_data.result = res
    return _buffered_probe_data.result

def _probe_daemon(daemon_port=None):
    daemon_url, headers = _daemon_url(daemon_port=daemon_port, no_client_auth=True)
    url = f'{daemon_url}/probe'
    try:
        x = _http_get_json(url)
    except Exception as e:
        return None
    res = _probe_result(x) if x is not None else None
    return res

def _kachery_offline_storage_dir_env_is_set():
    return os.getenv('KACHERY_OFFLINE_STORAGE_DIR', None) is not None

def _kachery_storage_dir():
    if _kachery_offline_storage_dir_env_is_set():
        return os.getenv('KACHERY_OFFLINE_STORAGE_DIR', None)
    else:
        p = _buffered_probe_daemon()
        if p is not None:
            return p.kachery_storage_dir
        else:
            return None

def _create_if_needed(dirpath: str) -> str:
    if not os.path.isdir(dirpath):
        try:
            os.mkdir(dirpath)
        except:
            # in case it was created elsewhere
            if not os.path.isdir(dirpath):
                raise Exception(f'Failed to create dir: {dirpath}')
    return dirpath

def _kachery_temp_dir() -> str:
    d = os.getenv('KACHERY_TEMP_DIR', None)
    if d is not None:
        return _create_if_needed(d)
    if _kachery_offline_storage_dir_env_is_set():
        return _create_if_needed(os.getenv('KACHERY_OFFLINE_STORAGE_DIR') + '/kachery-tmp')
    else:
        return _create_if_needed(tempfile.gettempdir() + '/kachery-tmp')

def _is_offline_mode():
    return _kachery_offline_storage_dir_env_is_set()
        
def _is_online_mode():
    if _is_offline_mode():
        return False
    return _kachery_storage_dir() is not None

def _get_node_id(daemon_port=None) -> str:
    x = _buffered_probe_daemon(daemon_port=daemon_port)
    assert x is not None, 'Unable to connect to daemon.'
    return x.node_id



import requests
from typing import Any, Union, cast
import json
from .._daemon_connection import _daemon_url
from .._misc import _http_post_json


def _update_task_status(*, channel: str, task_id: str, task_function_type: str, status: str, result: Union[Any, None]=None, error_message: Union[str, None]=None):
    query_result = None
    if status == 'finished':
        if task_function_type == 'pure-calculation':
            if result is None:
                raise Exception('No result provided for pure calculation even though status is finished')
            result_content = json.dumps(result).encode()
            signed_url = _create_signed_task_result_upload_url(channel=channel, task_id=task_id, size=len(result_content))
            _http_put_bytes(signed_url, result_content)
        elif task_function_type == 'query':
            if result is None:
                raise Exception('No result provided for query even though status is finished')
            query_result = result
    else:
        if result is not None:
            raise Exception('Result is not none even though status is not finished')
    
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/task/updateTaskStatus'
    # export interface TaskUpdateTaskStatusRequest {
    #     channelName: ChannelName
    #     taskId: TaskId,
    #     status: TaskStatus
    #     errorMessage?: ErrorMessage
    #     queryResult?: JSONValue
    # }
    req_data = {
        'channelName': channel,
        'taskId': task_id,
        'status': status
    }
    if query_result is not None:
        req_data['queryResult'] = query_result
    if error_message is not None:
        req_data['errorMessage'] = error_message
    x = _http_post_json(url, req_data, headers=headers)
    if not x['success']:
        raise Exception(f'Unable to update task status')

def _create_signed_task_result_upload_url(*, channel: str, task_id: str, size: int):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/task/createSignedTaskResultUploadUrl'
    # export type TaskCreateSignedTaskResultUploadUrlRequest = {
    #     channelName: ChannelName
    #     taskId: TaskId
    #     size: ByteCount
    # }
    req_data = {
        'channelName': channel,
        'taskId': task_id,
        'size': size
    }
    x = _http_post_json(url, req_data, headers=headers)
    # export type TaskCreateSignedTaskResultUploadUrlResponse = {
    #     success: boolean
    #     signedUrl: UrlString
    # }
    if not x['success']:
        raise Exception(f'Unable to create signed task result upload url')
    return cast(str, x['signedUrl'])

def _http_put_bytes(url: str, data: bytes):
    headers = {
        'Content-Type': 'application/octet-stream',
        'Content-Length': f'{len(data)}'
    }
    req = requests.put(url, data=data, headers=headers)
    try:
        if req.status_code != 200:
            raise Exception(f'Error putting data: {req.status_code} {req.content.decode("utf-8")}')
    finally:
        req.close()




import os
from typing import Any, Union
import subprocess
import simplejson
import numpy as np
import stat
import json
from ._daemon_connection import _is_offline_mode, _is_online_mode, _kachery_storage_dir, _daemon_url
from ._local_kachery_storage import _local_kachery_storage_store_file, _local_kachery_storage_link_file
from ._misc import _http_post_json, _http_post_file
from ._temporarydirectory import TemporaryDirectory
from ._safe_pickle import _safe_pickle, _safe_unpickle
from ._local_kachery_storage import _get_path_ext

def _store_file(path: str, basename: Union[str, None]=None) -> str:
    if basename is None:
        basename = os.path.basename(path)
    if _is_offline_mode():
        stored_path, hash0, manifest_hash = _local_kachery_storage_store_file(path=path)
        if manifest_hash is None:
            return f'sha1://{hash0}/{basename}'
        else:
            return f'sha1://{hash0}/{basename}?manifest={manifest_hash}'
    if not _is_online_mode():
        raise Exception('Not connected to daemon and not in offline mode.')
    file_size = os.path.getsize(path)
    daemon_url, headers = _daemon_url()
    # url = f'{daemon_url}/storeFile'
    url = f'{daemon_url}/store'
    headers['Content-Length'] = f'{file_size}'
    resp = _http_post_file(url, os.path.abspath(path), headers=headers)

    # resp = _http_post_json(url, {'localFilePath': os.path.abspath(path)}, headers=headers)
    if not resp['success']:
        raise Exception(f'Problem storing file: {resp["error"]}')
    sha1 = resp['sha1']
    manifest_sha1 = resp['manifestSha1']

    # important to verify that we can access the file
    # this is crucial for systems where the daemon is running on a different computer
    # in frank lab there was an issue where we needed to stat the file before proceeding
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    path0 = _get_path_ext(hash=sha1, create=False, directory=sha1_directory)
    if not os.path.exists(path0):
        raise Exception(f'Unexpected, could not find stored file after storing with daemon: {path0}')

    size0 = _get_file_size_using_system_call(path0)
    if size0 != file_size:
        if size0 == 0:
            # perhaps the file has not synced across devices
            raise Exception(f'Inconsistent size between stored file and original file for: {path} {path0} {file_size} {size0}')
        else:
            raise Exception(f'Unexpected size discrepancy between stored file and original file for: {path} {path0} {file_size} {size0}')

    if manifest_sha1:
        return f'sha1://{sha1}/{basename}?manifest={manifest_sha1}'
    else:
        return f'sha1://{sha1}/{basename}'

def _link_file(path: str, basename: Union[str, None]=None) -> str:
    if basename is None:
        basename = os.path.basename(path)
    if _is_offline_mode():
        stored_path, hash0, manifest_hash = _local_kachery_storage_link_file(path=path)
        if manifest_hash is None:
            return f'sha1://{hash0}/{basename}'
        else:
            return f'sha1://{hash0}/{basename}?manifest={manifest_hash}'
    if not _is_online_mode():
        raise Exception('Not connected to daemon and not in offline mode.')
    file_size = os.path.getsize(path)
    mtime = os.stat(path).st_mtime
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/linkFile'
    resp = _http_post_json(url, {'localFilePath': os.path.abspath(path), 'size': file_size, 'mtime': mtime}, headers=headers)

    if not resp['success']:
        raise Exception(f'Problem linking file: {resp["error"]}')
    sha1 = resp['sha1']
    manifest_sha1 = resp['manifestSha1']

    # important to verify that we can access the file
    # this is crucial for systems where the daemon is running on a different computer
    # in frank lab there was an issue where we needed to stat the file before proceeding
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    path0 = _get_path_ext(hash=sha1, create=False, directory=sha1_directory)
    if (not os.path.exists(path0)) and (not os.path.exists(path0 + '.link')):
        raise Exception(f'Unexpected, could not find stored file after linking with daemon: {path0}')

    if os.path.exists(path0):
        size0 = _get_file_size_using_system_call(path0)
        if size0 != file_size:
            if size0 == 0:
                # perhaps the file has not synced across devices
                raise Exception(f'Inconsistent size between stored/linked file and original file for: {path} {path0} {file_size} {size0}')
            else:
                raise Exception(f'Unexpected size discrepancy between stored/linked file and original file for: {path} {path0} {file_size} {size0}')
    else:
        try:
            with open(path0 + '.link', 'r') as f:
                json.load(f)
        except:
            raise Exception(f'Unexpected file reading link file after linking: {path}')

    if manifest_sha1:
        return f'sha1://{sha1}/{basename}?manifest={manifest_sha1}'
    else:
        return f'sha1://{sha1}/{basename}'

def _get_file_size_using_system_call(path: str):
    return int(subprocess.check_output(['stat', '-c%s', path]))

def _add_read_permissions(fname: str):
    st = os.stat(fname)
    os.chmod(fname, st.st_mode | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

def _add_exec_permissions(fname: str):
    st = os.stat(fname)
    os.chmod(fname, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

def _store_text(text: str, basename: Union[str, None]=None) -> str:
    if basename is None:
        basename = 'file.txt'
    with TemporaryDirectory() as tmpdir:
        fname = tmpdir + '/text.txt'
        with open(fname, 'w') as f:
            f.write(text)
        _add_read_permissions(tmpdir)
        _add_exec_permissions(tmpdir)
        _add_read_permissions(fname)
        return _store_file(fname, basename=basename)

def _store_json(object: Union[dict, list, int, float, str], basename: Union[str, None]=None, separators=(',', ':'), indent=None) -> str:
    if basename is None:
        basename = 'file.json'
    txt = simplejson.dumps(object, separators=separators, indent=indent)
    return _store_text(text=txt, basename=basename)

def _store_npy(array: np.ndarray, basename: Union[str, None]=None) -> str:
    if basename is None:
        basename = 'file.npy'
    with TemporaryDirectory() as tmpdir:
        fname = tmpdir + '/array.npy'
        np.save(fname, array, allow_pickle=False)
        _add_read_permissions(tmpdir)
        _add_exec_permissions(tmpdir)
        _add_read_permissions(fname)
        return _store_file(fname, basename=basename)

def _store_pkl(x: Any, basename: Union[str, None]=None) -> str:
    if basename is None:
        basename = 'file.pkl'
    with TemporaryDirectory() as tmpdir:
        fname = tmpdir + '/array.pkl'
        _safe_pickle(fname, x)
        _add_read_permissions(tmpdir)
        _add_exec_permissions(tmpdir)
        _add_read_permissions(fname)
        return _store_file(fname, basename=basename)


from typing import Any, Callable, Dict, List, Literal, Union

_global_registered_taskfunctions_by_function_id: Dict[str, Callable] = {}

def find_taskfunction(function_id: str) -> Union[Callable, None]:
    if function_id in _global_registered_taskfunctions_by_function_id:
        return _global_registered_taskfunctions_by_function_id[function_id]
    else:
        return None

def all_taskfunction_ids() -> List[str]:
    return list(_global_registered_taskfunctions_by_function_id.keys())

def taskfunction(function_id: str, *, type: Union[Literal['pure-calculation'], Literal['query'], Literal['action']]):
    def wrap(f: Callable[..., Any]):
        setattr(f, '_task_function_type', type)
        _global_registered_taskfunctions_by_function_id[function_id] = f
        return f
    return wrap


import sys
import os
import shutil
from typing import Any, Union
import simplejson
import numpy as np
from ._daemon_connection import _is_offline_mode, _is_online_mode, _daemon_url, _kachery_storage_dir
from ._misc import _create_file_key, _http_post_json_receive_json_socket, _parse_kachery_uri
from ._exceptions import LoadFileError
from ._local_kachery_storage import _local_kachery_storage_load_file, _local_kachery_storage_load_bytes
from ._safe_pickle import _safe_unpickle

def _load_file(uri: str, dest: Union[str, None]=None, local_only: bool=False) -> Union[str, None]:
    # handle old sha1dir system
    if uri.startswith('sha1dir://'):
        uri0 = _resolve_file_uri_from_dir_uri(uri)
        if uri0 is None:
            return None
        uri = uri0
    
    if not uri.startswith('sha1://'):
        if os.path.isfile(uri):
            local_path = uri
            if dest is not None:
                shutil.copyfile(local_path, dest)
                return dest
            else:
                return local_path
        else:
            raise Exception(f'Local file not found: {uri}')

    # first check the local kachery storage (if kachery storage dir is known)
    if _kachery_storage_dir():
        if True: # for debugging (not loading locally) switch to false
            protocol, algorithm, hash0, additional_path, query = _parse_kachery_uri(uri)
            if protocol != 'sha1':
                raise Exception(f'Protocol not supported: {protocol}')
            local_path = _local_kachery_storage_load_file(sha1_hash=hash0)
            if local_path is not None:
                if dest is not None:
                    shutil.copyfile(local_path, dest)
                    return dest
                else:
                    return local_path
    if _is_offline_mode():
        return None
    if not _is_online_mode():
        raise Exception('Not connected to daemon, and KACHERY_OFFLINE_STORAGE_DIR environment variable is not set.')
    
    if local_only:
        return None
    
    protocol, algorithm, hash0, additional_path, query = _parse_kachery_uri(uri)
    assert algorithm == 'sha1'
    file_key = _create_file_key(sha1=hash0, query=query)
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/loadFile'
    sock, req = _http_post_json_receive_json_socket(url, dict(
        fileKey=file_key
    ), headers=headers)
    try:
        for r in sock:
            try:
                type0 = r.get('type')
            except:
                raise Exception(f'Unexpected response from daemon: {r}: {uri}')
            if type0 == 'finished':
                print(f'Loaded file: {uri}')
                local_file_path: str = r['localFilePath']
                if not os.path.exists(local_file_path):
                    raise Exception(f'Unexpected in load_file: file does not exist: {local_file_path}')
                return local_file_path
            elif type0 == 'progress':
                bytes_loaded = r['bytesLoaded']
                bytes_total = r['bytesTotal']
                pct = (bytes_loaded / bytes_total) * 100
                print(f'Loaded {bytes_loaded} of {bytes_total} bytes ({pct:.1f} %): {uri}')
            elif type0 == 'error':
                return None
                # raise LoadFileError(f'Error loading file: {r["error"]}: {uri}')
            else:
                raise Exception(f'Unexpected message from daemon: {r}')
        # for url in _global_config['file_server_urls']:
        #     try:
        #         path = _load_file_from_file_server(uri=uri, dest=dest, file_server_url=url)
        #     except Exception as e:
        #         print(str(e))
        #         path = None
        #     if path:
        #         return path
        raise Exception(f'Unable to download file: {uri}')
    finally:
        req.close()

def _load_json(uri: str, local_only: bool=False) -> Union[dict, None]:
    local_path = _load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    with open(local_path, 'r') as f:
        return simplejson.load(f)

def _load_text(uri: str, local_only: bool=False) -> Union[str, None]:
    local_path = _load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    with open(local_path, 'r') as f:
        return f.read()

def _load_npy(uri: str, local_only: bool=False) -> Union[np.ndarray, Any, None]:
    local_path = _load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    return np.load(local_path, allow_pickle=False)

def _load_pkl(uri: str, local_only: bool=False) -> Union[np.ndarray, None]:
    local_path = _load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    return _safe_unpickle(local_path)

def _load_bytes(uri: str, start: Union[int, None], end: Union[int, None], write_to_stdout=False, local_only: bool=False) -> Union[bytes, None]:
    # handle old sha1dir system
    if uri.startswith('sha1dir://'):
        uri0 = _resolve_file_uri_from_dir_uri(uri)
        if uri0 is None:
            return None
        uri = uri0
    
    if not uri.startswith('sha1://'):
        if os.path.isfile(uri):
            local_path = uri
            return _load_bytes_from_local_file(local_path, start=start, end=end, write_to_stdout=write_to_stdout)
        else:
            raise Exception(f'Local file not found: {uri}')
    
    # first check the local kachery storage (if kachery storage dir is known)
    if _kachery_storage_dir():
        protocol, algorithm, hash0, additional_path, query = _parse_kachery_uri(uri)
        if protocol != 'sha1':
            raise Exception(f'Protocol not supported: {protocol}')
        bytes0 = _local_kachery_storage_load_bytes(sha1_hash=hash0, start=start, end=end, write_to_stdout=write_to_stdout)
        if bytes0 is not None:
            return bytes0
    
    if _is_offline_mode():
        return None
    if not _is_online_mode():
        raise Exception('Not connected to daemon, and KACHERY_OFFLINE_STORAGE_DIR environment variable is not set.')

    if local_only:
        return None
    
    protocol, algorithm, hash0, additional_path, query = _parse_kachery_uri(uri)
    if query.get('manifest'):
        manifest = _load_json(f'sha1://{query["manifest"][0]}')
        if manifest is None:
            print('Unable to load manifest')
            return None
        assert manifest['sha1'] == hash0, 'Manifest sha1 does not match expected.'
        data_chunks = []
        chunks_to_load = []
        for ch in manifest['chunks']:
            if start < ch['end'] and end > ch['start']:
                chunks_to_load.append(ch)
        for ii, ch in enumerate(chunks_to_load):
            if len(chunks_to_load) > 4:
                print(f'load_bytes: Loading chunk {ii + 1} of {len(chunks_to_load)}')
            chunk_uri = f'sha1://{ch["sha1"]}?chunkOf={hash0}~{ch["start"]}~{ch["end"]}'
            chunk_path = _load_file(chunk_uri)
            if chunk_path is None:
                print(f'Problem loading chunk: {chunk_uri}')
                return None
            start_byte = max(0, start - ch['start'])
            end_byte = min(ch['end']-ch['start'], end-ch['start'])
            a = _load_bytes(
                uri=chunk_path,
                start=start_byte,
                end=end_byte
            )
            if a is None:
                print(f'Unable to load bytes from chunk: {chunk_path} (start={start_byte}; end={end_byte})')
                return None
            data_chunks.append(a)
        return b''.join(data_chunks)
    
    path = _load_file(uri=uri)
    if path is None:
        print('Unable to load file.')
        return None
    bytes0 = _local_kachery_storage_load_bytes(sha1_hash=hash0, start=start, end=end, write_to_stdout=write_to_stdout)

def _load_bytes_from_local_file(local_fname: str, *, start: Union[int, None]=None, end: Union[int, None]=None, write_to_stdout: bool=False) -> Union[bytes, None]:
    size0 = os.path.getsize(local_fname)
    if start is None:
        start = 0
    if end is None:
        end = size0
    if start < 0 or start > size0 or end < start or end > size0:
        raise Exception('Invalid start/end range for file of size {}: {} - {}'.format(size0, start, end))
    if start == end:
        return bytes()
    with open(local_fname, 'rb') as f:
        f.seek(start)
        if write_to_stdout:
            ii = start
            while ii < end:
                nn = min(end - ii, 4096)
                data0 = f.read(nn)
                ii = ii + nn
                sys.stdout.buffer.write(data0)
            return None
        else:
            return f.read(end-start)

def _resolve_file_uri_from_dir_uri(dir_uri, local_only: bool=False):
    protocol, algorithm, hash0, additional_path, query = _parse_kachery_uri(dir_uri)
    assert protocol == algorithm + 'dir'
    dd = _load_json(algorithm + '://' + hash0, local_only=local_only)
    if dd is None:
        return None
    if additional_path:
        list0 = additional_path.split('/')
    else:
        list0 = []
    ii = 0
    while ii < len(list0):
        assert dd is not None
        name0 = list0[ii]
        if name0 in dd['dirs']:
            dd = dd['dirs'][name0]
        elif name0 in dd['files']:
            if ii + 1 == len(list0):
                hash1 = None
                algorithm1 = None
                for alg in ['sha1', 'md5']:
                    if alg in dd['files'][name0]:
                        hash1 = dd['files'][name0][alg]
                        algorithm1 = alg
                if hash1 is None:
                    return None
                return algorithm1 + '://' + hash1
            else:
                return None
        else:
            return None
        ii = ii + 1
    return None


import hashlib
import json
import time
from os import wait
from os.path import basename
from typing import Union
from urllib.parse import quote, unquote

from ._load_file import _load_json
from ._store_file import _store_json
from ._daemon_connection import _daemon_url
from ._misc import _http_post_json, _http_get_json


class Feed:
    def __init__(self, uri, *, timeout_sec: Union[None, float]=None):
        if '://' not in uri:
            uri = f'feed://{uri}'
        self._feed_uri = uri
        self._timeout_sec = timeout_sec
        if uri.startswith('feed://'):
            feed_id, subfeed_name, position = _parse_feed_uri(uri)
            if subfeed_name is not None:
                raise Exception('Cannot specify subfeed name in URI when loading feed')
            self._feed_id = feed_id
            self._is_writeable = None
            self._is_snapshot = False
            self._initialize()
        elif uri.startswith('sha1://'):
            # snapshot
            self._feed_id = None
            self._is_writeable = False
            self._is_snapshot = True
            self._snapshot_object = _load_json(uri)
            assert self._snapshot_object is not None, f'Unable to load snapshot: {uri}'
        else:
            raise Exception(f'Unexpected feed uri: {uri}')
    def _initialize(self):
        daemon_url, headers = _daemon_url()
        url = f'{daemon_url}/feed/getFeedInfo'
        x = _http_post_json(url, dict(
            feedId=self._feed_id
        ), headers=headers)

        assert x['success'], f'Unable to initialize feed: {self._feed_id} ({x["error"]})'
        self._is_writeable = x['isWriteable']
    @property
    def is_writeable(self):
        return self._is_writeable
    @property
    def feed_id(self):
        return self._feed_id
    @property
    def uri(self):
        return self._feed_uri
    @property
    def is_snapshot(self):
        return self._is_snapshot
    def load_subfeed(self, subfeed_name, position=0):
        return Subfeed(feed=self, subfeed_name=subfeed_name, position=position)
    def delete(self):
        _delete_feed(self.uri)
    def create_snapshot(self, subfeed_names: list):
        subfeeds = dict()
        for subfeed_name in subfeed_names:
            subfeed = self.load_subfeed(subfeed_name)
            messages = subfeed.get_next_messages(wait_msec=0)
            subfeeds[subfeed.subfeed_hash] = dict(
                subfeedHash=subfeed.subfeed_hash,
                messages=messages
            )
        snapshot_uri = _store_json(dict(
            subfeeds=subfeeds
        ), basename='feed.json')
        return Feed(snapshot_uri)

def _subfeed_hash(subfeed_name):
    if isinstance(subfeed_name, str):
        if subfeed_name.startswith('~'):
            return subfeed_name[1:]
        return _sha1_of_string(subfeed_name)
    else:
        return _sha1_of_object(subfeed_name)

def _sha1_of_string(txt: str) -> str:
    hh = hashlib.sha1(txt.encode('utf-8'))
    ret = hh.hexdigest()
    return ret

def _sha1_of_object(obj: object) -> str:
    txt = json.dumps(obj, sort_keys=True, separators=(',', ':'))
    return _sha1_of_string(txt)


class Subfeed:
    def __init__(self, *, feed, subfeed_name, position):
        self._feed = feed
        self._feed_uri = feed._feed_uri
        self._feed_id = feed._feed_id
        self._is_writeable = feed._is_writeable
        self._subfeed_name = subfeed_name
        self._position = position
        self._subfeed_hash = _subfeed_hash(self._subfeed_name)

        if isinstance(self._subfeed_name, str):
            self._subfeed_name_str = self._subfeed_name
        else:
            self._subfeed_name_str = '~' + self._subfeed_hash

        self._initialize()

    def _initialize(self):
        pass

    @property
    def uri(self):
        feed_uri = self._feed_uri
        if feed_uri.startswith('feed://'):
            return f'{self._feed_uri}/{quote(self._subfeed_name_str)}'
        elif feed_uri.startswith('sha1://'):
            return f'{self._feed_uri}?subfeedName={quote(self._subfeed_name_str)}'
        else:
            raise Exception(f'Unexpected feed uri: {feed_uri}')

    @property
    def position(self):
        return self._position
    
    @property
    def subfeed_name(self):
        return self._subfeed_name
    
    @property
    def subfeed_hash(self):
        return self._subfeed_hash

    def set_position(self, position):
        self._position = position

    def get_num_local_messages(self):
        if not self.is_snapshot:
            daemon_url, headers = _daemon_url()
            url = f'{daemon_url}/feed/getNumLocalMessages'
            x = _http_post_json(url, dict(
                feedId=self._feed_id,
                subfeedHash=self._subfeed_hash
            ), headers=headers)
            assert x['success'], f'Unable to get num. messages for subfeed: {self._feed_id} {self._subfeed_name_str}'
            return x['numMessages']
        else:
            messages = self._get_snapshot_messages()
            return len(messages)
        
    def _get_snapshot_messages(self):
        # only applies when feed is a snapshot
        try:
            obj = self._feed._snapshot_object['subfeeds'][self._subfeed_hash]
            return obj['messages']
        except:
            return []

    def get_next_messages(self, *, wait_msec=10, signed=False, max_num_messages=0, advance_position=True):
        if not self.is_snapshot:
            subfeed_watches = {
                'watch': {
                    'feedId': self._feed_id,
                    'subfeedHash': self._subfeed_hash,
                    'position': self._position
                }
            }
            x = _watch_for_new_messages(subfeed_watches, wait_msec=wait_msec, signed=signed, max_num_messages=max_num_messages)
            y = x.get('watch', [])
            if advance_position:
                self._position = self._position + len(y)
            return y
        else:
            messages = self._get_snapshot_messages()
            position = self._position
            if max_num_messages > 0:
                ret = messages[position:position + max_num_messages]
            else:
                ret = messages[position:]
            if advance_position:
                self._position = self._position + len(ret)
            return ret

    def get_next_message(self, *, wait_msec, signed=False, advance_position=True):
        messages = self.get_next_messages(wait_msec=wait_msec, signed=signed, max_num_messages=1, advance_position=advance_position)
        if messages is None:
            return None
        if len(messages) == 0:
            return None
        return messages[0]
    
    def message_stream(self, *, signed=False):
        class custom_iterator:
            def __init__(self, parent):
                self._parent = parent
                self._messages = []
                self._relative_position = 0
                self._load_messages()
            
            def _load_messages(self):
                messages = self._parent.get_next_messages(wait_msec=5000, signed=signed, advance_position=False)
                if messages is None:
                    return
                for msg in messages:
                    self._messages.append(msg)

            def __iter__(self):
                return self

            def __next__(self):
                while self._relative_position >= len(self._messages):
                    self._load_messages()
                    if self._relative_position >= len(self._messages):
                        if self._parent.is_snapshot:
                            raise StopIteration
                        time.sleep(0.05)
                self._parent._position = self._parent._position + 1
                self._relative_position = self._relative_position + 1
                return self._messages[self._relative_position - 1]
        return custom_iterator(parent=self)
    
    @property
    def is_snapshot(self):
        return self._feed.is_snapshot
    
    @property
    def is_writeable(self):
        return self._feed.is_writeable

    def print_messages(self):
        for msg in self.message_stream():
            print(msg)
    
    def print_signed_messages(self):
        for msg in self.message_stream(signed=True):
            print(msg)

    def append_message(self, message):
        self.append_messages([message])

    def append_messages(self, messages):
        if not self.is_writeable:
            raise Exception('Cannot append messages to a readonly feed')
        # CHAIN:append_messages:step(1)
        daemon_url, headers = _daemon_url()
        url = f'{daemon_url}/feed/appendMessages'
        x = _http_post_json(url, dict(
            feedId=self._feed_id,
            subfeedHash=self._subfeed_hash,
            messages=messages
        ), headers=headers)
        if not x['success']:
            raise Exception(f'Unable to append messages: {x.get("error")}')

def _create_feed(feed_name=None):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/feed/createFeed'
    req_data = dict()
    if feed_name is not None:
        req_data['feedName'] = feed_name
    x = _http_post_json(url, req_data, headers=headers)
    if not x['success']:
        raise Exception(f'Unable to create feed: {feed_name}')
    return _load_feed('feed://' + x['feedId'])

def _delete_feed(feed_name_or_uri):
    if feed_name_or_uri.startswith('feed://'):
        feed_uri = feed_name_or_uri
        feed_id, subfeed_name, position = _parse_feed_uri(feed_uri)
        assert subfeed_name is None, 'Cannot specify subfeed name when deleting feed'
        daemon_url, headers = _daemon_url()
        url = f'{daemon_url}/feed/deleteFeed'
        x = _http_post_json(url, dict(
            feedId=feed_id
        ), headers=headers)
        if not x['success']:
            raise Exception(f'Unable to delete feed {feed_id}: {x.get("error", None)}')
    else:
        feed_name = feed_name_or_uri
        feed_id = _get_feed_id(feed_name, create=False)
        assert feed_id is not None, f'Unable to find feed with name: {feed_name}'
        _delete_feed(f'feed://{feed_id}')

def _get_feed_id(feed_name, *, create=False):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/feed/getFeedId'
    x = _http_post_json(url, dict(
        feedName=feed_name
    ), headers=headers)
    if not x['success']:
        if create:
            return _create_feed(feed_name)._feed_id
        else:
            raise Exception(f'Unable to load feed with name: {feed_name}')
    feed_id = x['feedId']
    return feed_id

def _load_subfeed(subfeed_uri):
    feed_id, subfeed_name, position = _parse_feed_uri(subfeed_uri)
    assert subfeed_name is not None, 'No subfeed name found'
    return Feed('feed://' + feed_id).load_subfeed(subfeed_name=subfeed_name, position=position)
        
def _load_feed(feed_name_or_uri, *, timeout_sec: Union[None, float]=None, create=False):
    if feed_name_or_uri.startswith('feed://'):
        if create is True:
            raise Exception('Cannot use create=True when feed ID is specified.')
        feed_uri = feed_name_or_uri
        feed_id, subfeed_name, position = _parse_feed_uri(feed_uri)
        assert subfeed_name is None, 'Not a feed uri'
        return Feed('feed://' + feed_id, timeout_sec=timeout_sec)
    elif feed_name_or_uri.startswith('sha1://'):
        if create is True:
            raise Exception('Cannot use create=True when feed is a snapshot.')
        feed_uri = feed_name_or_uri
        return Feed(feed_uri)
    else:
        feed_name = feed_name_or_uri
        feed_id = _get_feed_id(feed_name, create=create)
        return _load_feed(f'feed://{feed_id}')

def _watch_for_new_messages(subfeed_watches, *, wait_msec, signed=False, max_num_messages=0):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/feed/watchForNewMessages'
    subfeed_watches2 = {}
    for key, watch in subfeed_watches.items():
        subfeed_watches2[key] = {
            'feedId': watch['feedId'],
            'subfeedHash': watch.get('subfeedHash') if 'subfeedHash' in watch else _subfeed_hash(watch['subfeedName']),
            'position': watch['position']
        }
    x = _http_post_json(url, dict(
        subfeedWatches=subfeed_watches2,
        waitMsec=wait_msec,
        signed=signed,
        maxNumMessages=max_num_messages
    ), headers=headers)
    if not x['success']:
        raise Exception(f'Unable to watch for new messages.')
    return x['messages']

def _parse_feed_uri(uri):
    listA = uri.split('?')
    assert len(listA) >= 1
    list0 = listA[0].split('/')
    assert len(list0) >= 3
    protocol = list0[0].replace(':', '')
    assert protocol == 'feed'
    feed_id = list0[2]
    if len(list0) >= 3:
        subfeed_name = '/'.join(list0[3:])
    else:
        subfeed_name = None
    if subfeed_name:
        subfeed_name = unquote(subfeed_name)
    else:
        subfeed_name = None
    return feed_id, subfeed_name, 0



import signal

# Thanks: https://stackoverflow.com/questions/842557/how-to-prevent-a-block-of-code-from-being-interrupted-by-keyboardinterrupt-in-py
class PreventKeyboardInterrupt(object):
    def __enter__(self):
        self.signal_received = False
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig, frame):
        self.signal_received = (sig, frame)
        print('SIGINT received. Preventing KeyboardInterrupt.')

    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, self.old_handler)


from .version import __version__
from .main import load_bytes, load_feed, load_file, load_json, load_npy, load_pkl, load_subfeed, load_text
from .main import create_feed
from .main import store_file, store_json, store_npy, store_pkl, store_text, link_file
from .main import get, set, get_feed_id, get_string
from .main import watch_for_new_messages

from .request_task import request_task
from .task_backend.taskfunction import taskfunction
from .task_backend.run_task_backend import run_task_backend

from ._daemon_connection import _kachery_storage_dir, _kachery_temp_dir
from ._temporarydirectory import TemporaryDirectory
from ._shellscript import ShellScript

from .cli import cli


from typing import Any, Dict, List
from .RequestedTask import RequestedTask

class TaskJob:
    def __init__(self, *, requested_task: RequestedTask, job: Any) -> None:
        import hither2 as hi
        if not isinstance(job, hi.Job):
            raise Exception('Not a hither job.')
        self._requested_task = requested_task
        self._job = job
    @property
    def requested_task(self):
        return self._requested_task
    @property
    def job(self):
        return self._job

class TaskJobManager:
    def __init__(self):
        self._task_jobs: Dict[str, TaskJob] = {}
    def add_task_job(self, *, requested_task: RequestedTask, job: Any):
        import hither2 as hi # Only import hither if we have a hither job
        if not isinstance(job, hi.Job):
            raise Exception('Not a hither job.')
        if requested_task.task_id in self._task_jobs:
            raise Exception('Unexpected. Already have job with task hash.')
        self._task_jobs[requested_task.task_id] = TaskJob(requested_task=requested_task, job=job)
    def get_existing_job_for_task(self, requested_task: RequestedTask):
        if requested_task.task_id in self._task_jobs:
            return self._task_jobs[requested_task.task_id].job
        else:
            return None
    def process_events(self):
        task_jobs_to_delete: List[str] = []
        has_jobs = False
        for k, v in self._task_jobs.items():
            has_jobs = True
            job = v.job
            requested_task = v.requested_task
            if job.status != requested_task.status:
                if job.status == 'error':
                    error_message = str(job.result.error)
                else:
                    error_message = None
                if job.status == 'finished':
                    result = job.result.return_value
                else:
                    result = None
                requested_task.update_status(status=job.status, error_message=error_message, result=result)
            if requested_task.status in ['error', 'finished']:
                task_jobs_to_delete.append(k)
        for k in task_jobs_to_delete:
            del self._task_jobs[k]
        if has_jobs:
            import hither2 as hi # Only import hither if we have a hither job
            hi.wait(0)
            


import os
import time
import json
from typing import Dict, Iterable, Optional, Tuple, Union
from urllib.parse import parse_qs

from requests.models import Response

def _parse_kachery_uri(uri: str) -> Tuple[str, str, str, str, dict]:
    listA = uri.split('?')
    if len(listA) > 1:
        query = parse_qs(listA[1])
    else:
        query = {}
    list0 = listA[0].split('/')
    protocol = list0[0].replace(':', '')
    hash0 = list0[2]
    if '.' in hash0:
        hash0 = hash0.split('.')[0]
    additional_path = '/'.join(list0[3:])
    algorithm = None
    for alg in ['sha1', 'md5', 'key']:
        if protocol.startswith(alg):
            algorithm = alg
    if algorithm is None:
        raise Exception('Unexpected protocol: {}'.format(protocol))
    return protocol, algorithm, hash0, additional_path, query

def _http_post_json(url: str, data: dict, verbose: Optional[bool] = None, headers: dict = {}) -> dict:
    timer = time.time()
    if verbose is None:
        verbose = (os.environ.get('HTTP_VERBOSE', '') == 'TRUE')
    if verbose:
        print('_http_post_json::: ' + url)
    try:
        import requests
    except:
        raise Exception('Error importing requests *')
    req = requests.post(url, json=data, headers=headers)
    try:
        if req.status_code != 200:
            return dict(
                success=False,
                error='Error posting json: {} {}'.format(
                    req.status_code, req.content.decode('utf-8'))
            )
        if verbose:
            print('Elapsed time for _http_post_json: {}'.format(time.time() - timer))
        return json.loads(req.content)
    finally:
        req.close()

def _http_post_json_receive_json_socket(url: str, data: dict, verbose: Optional[bool] = None, headers: dict = {}) -> Tuple[Iterable[dict], Response]:
    timer = time.time()
    if verbose is None:
        verbose = (os.environ.get('HTTP_VERBOSE', '') == 'TRUE')
    if verbose:
        print('_http_post_json::: ' + url)
    try:
        import requests
    except:
        raise Exception('Error importing requests *')
    req = requests.post(url, json=data, stream=True, headers=headers)
    if req.status_code != 200:
        raise Exception('Error posting json: {} {}'.format(req.status_code, req.content.decode('utf-8')))
    class custom_iterator:
        def __init__(self):
            pass

        def __iter__(self):
            return self

        def __next__(self):
            buf = bytearray(b'')
            while True:
                c = req.raw.read(1)
                if len(c) == 0:
                    raise StopIteration
                if c == b'#':
                    size = int(buf)
                    x = req.raw.read(size)
                    obj = json.loads(x)
                    return obj
                else:
                    buf.append(c[0])
    return custom_iterator(), req

def _http_get_json(url: str, verbose: Optional[bool] = None, headers: dict = {}) -> dict:
    timer = time.time()
    if verbose is None:
        verbose = (os.environ.get('HTTP_VERBOSE', '') == 'TRUE')
    if verbose:
        print('_http_get_json::: ' + url)
    try:
        import requests
    except:
        raise Exception('Error importing requests *')
    req = requests.get(url, headers=headers)
    try:
        if req.status_code != 200:
            return dict(
                success=False,
                error='Error getting json: {} {}'.format(
                    req.status_code, req.content.decode('utf-8'))
            )
        if verbose:
            print('Elapsed time for _http_get_json: {}'.format(time.time() - timer))
        return json.loads(req.content)
    finally:
        req.close()

def _http_post_file(url: str, file_path: str, headers: dict = {}) -> dict:
    try:
        import requests
    except:
        raise Exception('Error importing requests *')
    with open(file_path, 'rb') as f:
        req = requests.post(url, data=f, headers=headers)
    try:
        if req.status_code != 200:
            raise Exception(f'Error posting file: {url} {file_path}')
        return json.loads(req.content)
    finally:
        req.close()

def _create_file_key(*, sha1, query):
    file_key: Dict[str, Union[str, dict]] = dict(
        sha1=sha1
    )
    if 'manifest' in query:
        file_key['manifestSha1'] = query['manifest'][0]
    if 'chunkOf' in query:
        v = query['chunkOf'][0].split('~')
        assert len(v) ==3, 'Unexpected chunkOf in URI query.'
        file_key['chunkOf'] = {
            'fileKey': {
                'sha1': v[0]
            },
            'startByte': int(v[1]),
            'endByte': int(v[2])
        }
    return file_key


import json
import time
from typing import List
from .RequestedTask import RequestedTask
from os import error
from .taskfunction import find_taskfunction, all_taskfunction_ids
from .RegisteredTaskFunction import RegisteredTaskFunction
from .TaskBackend import TaskBackend
from .TaskJobManager import TaskJobManager


def run_task_backend(*, channels: List[str], task_function_ids: List[str]):
    task_functions: List[RegisteredTaskFunction] = []
    for function_id in task_function_ids:
        f = find_taskfunction(function_id)
        if f is None:
            raise Exception(f'Task function not found: {function_id}')
        for channel in channels:
            print(f'Registering task for channel {channel}: {function_id} ({f._task_function_type})')
            x = RegisteredTaskFunction(task_function_id=function_id, task_function_type=f._task_function_type, channel=channel)
            task_functions.append(x)
    function_ids_not_included = [id for id in all_taskfunction_ids() if id not in task_function_ids]
    if len(function_ids_not_included):
        print('WARNING: The following task functions are not registered:')
        print(', '.join([f"'{x}'" for x in function_ids_not_included]))
        print('')

    B = TaskBackend(registered_task_functions=task_functions)
    B.start()
    try:
        while True:
            B.process_events()
            time.sleep(0.1)
    finally:
        B.stop()



from typing import Any, Dict, Iterable, List, Union

import numpy as np

from ._feeds import (_create_feed, _delete_feed, _get_feed_id, _load_feed,
                     _load_subfeed, _watch_for_new_messages)
from ._mutables import (_get, _set, _delete)

from ._load_file import _load_file, _load_bytes, _load_text, _load_json, _load_npy, _load_pkl
from ._store_file import _store_file, _store_text, _store_json, _store_npy, _store_pkl, _link_file
from ._daemon_connection import _get_node_id

def load_file(
    uri: str,
    dest: Union[str, None]=None
) -> Union[str, None]:
    """Load a file either from local kachery storage or from a remote kachery node

    Args:
        uri (str): The kachery URI for the file to load: sha1://...
        dest (Union[str, None], optional): Optional location to copy the file to. Defaults to None.

    Returns:
        Union[str, None]: If found, the local path of the loaded file, else None
    """
    return _load_file(uri=uri, dest=dest)

def load_bytes(uri: str, start: int, end: int, write_to_stdout=False) -> Union[bytes, None]:
    """Load a subset of bytes from a file in local storage or from remote nodes in the kachery network

    Args:
        uri (str): The kachery URI for the file to load: sha1://...
        start (int): The start byte (inclusive)
        end (int): The end byte (not inclusive)

    Returns:
        Union[bytes, None]: The bytes if found, else None
    """
    return _load_bytes(uri=uri, start=start, end=end, write_to_stdout=write_to_stdout)

def load_json(uri: str) -> Union[dict, None]:
    """Load an object (Python dict) either from local kachery storage or from a remote kachery node

    Args:
        uri (str): The kachery URI for the file to load: sha1://...

    Returns:
        Union[dict, None]: If found, the Python dict, else None
    """
    return _load_json(uri=uri)

def load_text(uri: str) -> Union[str, None]:
    """Load a text string either from local kachery storage or from a remote kachery node

    Args:
        uri (str): The kachery URI for the file to load: sha1://...

    Returns:
        Union[str, None]: If found, the text string, else None
    """
    return _load_text(uri=uri)

def load_npy(uri: str) -> Union[np.ndarray, None]:
    """Load a Numpy array either from local kachery storage or from a remote kachery node

    Args:
        uri (str): The kachery URI for the .npy file to load: sha1://...

    Returns:
        Union[str, None]: If found, the Numpy array, else None
    """
    return _load_npy(uri=uri)

def load_pkl(uri: str) -> Union[Any, None]:
    """Load a Python item from a restricted pickle format either from local kachery storage or from a remote kachery node

    Args:
        uri (str): The kachery URI for the .pkl file to load: sha1://...

    Returns:
        Union[str, None]: If found, result, else None
    """
    return _load_pkl(uri=uri)

def store_file(path: str, basename: Union[str, None]=None) -> str:
    """Store file in the local kachery storage (will therefore be available on the kachery network) and return a kachery URI

    Args:
        path (str): Path of the file to store
        basename (Union[str, None], optional): Optional base file name to append to the sha1:// URI. Defaults to None.

    Returns:
        str: The kachery URI: sha1://...
    """
    return _store_file(path=path, basename=basename)

def link_file(path: str, basename: Union[str, None]=None) -> str:
    """Link a local file in the local kachery storage (will therefore be available on the kachery network) and return a kachery URI

    Args:
        path (str): Path of the file to link
        basename (Union[str, None], optional): Optional base file name to append to the sha1:// URI. Defaults to None.

    Returns:
        str: The kachery URI: sha1://...
    """
    return _link_file(path=path, basename=basename)

def store_json(object: Union[dict, list, int, float, str], basename: Union[str, None]=None) -> str:
    """Store object (Python dict, list or other jsonable) in the local kachery storage (will therefore be available on the kachery network) and return a kachery URI

    Args:
        object (dict): The Python dict to store
        basename (Union[str, None], optional): Optional base file name to append to the sha1:// URI. Defaults to None.

    Returns:
        str: The kachery URI: sha1://...
    """
    return _store_json(object=object, basename=basename)

def store_text(text: str, basename: Union[str, None]=None) -> str:
    """Store text in the local kachery storage (will therefore be available on the kachery network) and return a kachery URI

    Args:
        text (str): The text string to store
        basename (Union[str, None], optional): Optional base file name to append to the sha1:// URI. Defaults to None.

    Returns:
        str: The kachery URI: sha1://...
    """
    return _store_text(text=text, basename=basename)

def store_npy(array: np.ndarray, basename: Union[str, None]=None) -> str:
    """Store Numpy array in the local kachery storage (will therefore be available on the kachery network) and return a kachery URI

    Args:
        array (np.ndarray): The Numpy array to store
        basename (Union[str, None], optional): Optional base file name to append to the sha1:// URI. Defaults to None.

    Returns:
        str: The kachery URI: sha1://...
    """
    return _store_npy(array=array, basename=basename)

def store_pkl(x: Any, basename: Union[str, None]=None) -> str:
    """Store Python item in the local kachery storage using a restricted pickle format and return a kachery URI

    Args:
        array (Any): The item to store
        basename (Union[str, None], optional): Optional base file name to append to the sha1:// URI. Defaults to None.

    Returns:
        str: The kachery URI: sha1://...
    """
    return _store_pkl(x=x, basename=basename)

def get_node_id(daemon_port=None) -> str:
    """Return the Node ID for this kachery node

    Args:
        daemon_port ([type], optional): Optionally override the port of the API daemon. Defaults to None.

    Returns:
        str: The node ID
    """
    return _get_node_id(daemon_port=daemon_port)

################################################

def create_feed(feed_name: Union[str, None]=None):
    """Create a new local feed and optionally associate it with a local name

    Args:
        feed_name (Union[str, None], optional): Optional name for local retrieval. Defaults to None.

    Returns:
        Feed: The newly-created local writeable feed
    """
    return _create_feed(feed_name=feed_name)

def delete_feed(feed_name_or_uri: str) -> None:
    """Delete a feed with a particular name or URI

    Args:
        feed_name_or_uri (str): The name or URI of the feed to delete
    """
    return _delete_feed(feed_name_or_uri=feed_name_or_uri)    

def get_feed_id(feed_name: str, *, create: bool=False) -> Union[None, str]:
    """Return the ID of a feed by name, and optionally create a new feed if it does not exist

    Args:
        feed_name (str): The local name of the feed ID to retrieve
        create (bool, optional): Whether to create a new feed if none exists with this name. Defaults to False.

    Returns:
        Union[None, str]: The feed ID, if exists, else None
    """
    return _get_feed_id(feed_name=feed_name, create=create)

def load_subfeed(subfeed_uri: str):
    """Load a subfeed by URI

    Args:
        subfeed_uri (str): the URI of the subfeed

    Returns:
        Subfeed: The subfeed associated with the URI
    """
    return _load_subfeed(subfeed_uri=subfeed_uri)
        
def load_feed(feed_name_or_uri: str, *, timeout_sec: Union[None, float]=None, create=False):
    """Load a feed by URI or local name

    Args:
        feed_name_or_uri (str): Either the local name or the URI of the feed to load
        timeout_sec (Union[None, float], optional): An optional timeout for searching for the feed. Defaults to None.
        create (bool, optional): Whether to create if doesn't exist (only applies when supplying a local feed name). Defaults to False.

    Returns:
        Feed: The loaded feed
    """
    return _load_feed(feed_name_or_uri=feed_name_or_uri, timeout_sec=timeout_sec, create=create)

def watch_for_new_messages(subfeed_watches: Dict[str, dict], *, wait_msec, signed=False) -> Dict[str, Any]:
    """Watch for new messages on one or more subfeeds

    Args:
        subfeed_watches (List[dict]): A list of subfeed watches (TODO: more details needed)
        wait_msec ([type]): The wait duration for retrieving the messages

    Returns:
        List[dict]: The list of retrieved messages (TODO: more details needed)
    """
    return _watch_for_new_messages(subfeed_watches=subfeed_watches, wait_msec=wait_msec, signed=signed)

################################################

def set(key: Union[str, dict, list], value: Union[str, dict, list]):
    """Set a mutable value (only available locally)

    Args:
        key (Union[str, dict, list]): The key
        value (Union[str, dict, list]): The value

    Returns:
        None
    """
    return _set(key=key, value=value)

def get(key: Union[str, dict, list]):
    """Get a mutable value (only available locally)

    Args:
        key (Union[str, dict, list]): The key

    Returns:
        value (Union[str, dict, list, None]): The value if found, else None
    """
    return _get(key=key)

def get_string(key: Union[str, dict, list]):
    """Get a mutable value as a string (only available locally)

    Args:
        key (Union[str, dict, list]): The key

    Returns:
        value (Union[str, None]): The value as a string if found, else None
    """
    v = get(key)
    return str(v) if v is not None else None

def delete(key: Union[str, dict, list], value: Union[str, dict, list]):
    """Delete a mutable value (only available locally)

    Args:
        key (Union[str, dict, list]): The key

    Returns:
        None
    """
    return _delete(key=key)


import subprocess
import tempfile
import shutil
import signal
import os
import time
import random
import atexit
from typing import Optional, List, Any
from ._preventkeyboardinterrupt import PreventKeyboardInterrupt
from ._daemon_connection import _kachery_temp_dir

_running_scripts = {}

class ShellScript():
    def __init__(self, script: str, script_path: Optional[str]=None, keep_temp_files: bool=False, verbose: bool=False, label='', redirect_output_to_stdout=False):
        self._script_path = script_path
        self._keep_temp_files = keep_temp_files
        self._process: Optional[subprocess.Popen] = None
        self._files_to_remove: List[str] = []
        self._dirs_to_remove: List[str] = []
        self._start_time: Optional[float] = None
        self._verbose = verbose
        self._label = label
        self._redirect_output_to_stdout = redirect_output_to_stdout
        self._script_id = _random_string(10)

        lines = script.splitlines()
        lines = self._remove_initial_blank_lines(lines)
        if len(lines) > 0:
            num_initial_spaces = self._get_num_initial_spaces(lines[0])
            for ii, line in enumerate(lines):
                if len(line.strip()) > 0:
                    n = self._get_num_initial_spaces(line)
                    if n < num_initial_spaces:
                        print(script)
                        raise Exception('Problem in script. First line must not be indented relative to others')
                    lines[ii] = lines[ii][num_initial_spaces:]
        self._script = '\n'.join(lines)

    def substitute(self, old: str, new: Any) -> None:
        self._script = self._script.replace(old, '{}'.format(new))

    def write(self, script_path: Optional[str]=None) -> None:
        if script_path is None:
            script_path = self._script_path
        if script_path is None:
            raise Exception('Cannot write script. No path specified')
        with open(script_path, 'w', newline='\n') as f:
            f.write(self._script)
        os.chmod(script_path, 0o744)

    def start(self) -> None:
        if self._script_path is not None:
            script_path = self._script_path
        else:
            tempdir = tempfile.mkdtemp(prefix='tmp_shellscript_', dir=_kachery_temp_dir())
            script_path = os.path.join(tempdir, 'script.sh')
            self._dirs_to_remove.append(tempdir)        
        self.write(script_path)
        cmd = script_path
        if self._verbose:
            print('RUNNING SHELL SCRIPT: ' + cmd)
        self._start_time = time.time()
        _running_scripts[self._script_id] = self
        if self._redirect_output_to_stdout:
            self._process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        else:
            self._process = subprocess.Popen(cmd)

    def wait(self, timeout=None) -> Optional[int]:
        timeout_increment = 0.01
        if timeout is None or (timeout > timeout_increment):
            timer = time.time()
            while True:
                retcode = self.wait(timeout=timeout_increment)
                if retcode is not None:
                    self._cleanup()
                    return retcode
                if timeout is not None:
                    elapsed = time.time() - timer
                    if elapsed > timeout:
                        return None

        if not self.isRunning():
            self._print_stdout()
            self._cleanup()
            return self.returnCode()
        assert self._process is not None, "Unexpected self._process is None even though it is running."
        try:
            retcode = self._process.wait(timeout=timeout)
        except:
            retcode = None
        finally:
            self._print_stdout()
        if retcode is not None:
            self._cleanup()
        return retcode
    
    def _print_stdout(self):
        if not self._redirect_output_to_stdout:
            return
        if self._process is None:
            return
        for line in self._process.stdout:
            if isinstance(line, bytes):
                line = line.decode('utf-8')
            print(line)

    def _cleanup(self) -> None:
        try:
            if not hasattr(self, '_dirs_to_remove'):
                return
            if self._keep_temp_files:
                return
            for dirpath in self._dirs_to_remove:
                _rmdir_with_retries(dirpath, num_retries=5)
        except:
            print('WARNING: Problem in cleanup() of ShellScript')
        finally:
            if self._script_id in _running_scripts:
                del _running_scripts[self._script_id]

    def stop(self) -> None:
        with PreventKeyboardInterrupt():
            if not self.isRunning():
                return
            assert self._process is not None, "Unexpected self._process is None even though it is running."

            try:
                signals = [signal.SIGINT, signal.SIGINT, signal.SIGINT] + [signal.SIGTERM, signal.SIGTERM, signal.SIGTERM] + [signal.SIGKILL]
                signal_strings = ['SIGINT', 'SIGINT', 'SIGINT'] + ['SIGTERM', 'SIGTERM', 'SIGTERM'] + ['SIGKILL']
                delays = [5, 5, 5] + [5, 5, 5] + [1]

                for iis in range(len(signals)):
                    self._process.send_signal(signals[iis])
                    try:
                        self._process.wait(timeout=delays[iis])
                        return
                    except:
                        pass
            finally:
                self._cleanup()

    def kill(self) -> None:
        if not self.isRunning():
            return
        
        assert self._process is not None, "Unexpected self._process is None even though it is running."
        self._process.send_signal(signal.SIGKILL)
        try:
            self._process.wait(timeout=1)
        except:
            print('WARNING: unable to kill shell script.')
            pass

    def stopWithSignal(self, sig, timeout) -> bool:
        if not self.isRunning():
            return True
        
        assert self._process is not None, "Unexpected self._process is None even though it is running."
        self._process.send_signal(sig)
        try:
            self._process.wait(timeout=timeout)
            return True
        except:
            return False

    def elapsedTimeSinceStart(self) -> Optional[float]:
        if self._start_time is None:
            return None
        
        return time.time() - self._start_time

    def isRunning(self) -> bool:
        if not self._process:
            return False
        retcode = self._process.poll()
        if retcode is None:
            return True
        return False

    def isFinished(self) -> bool:
        if not self._process:
            return False
        return not self.isRunning()

    def returnCode(self) -> Optional[int]:
        if not self.isFinished():
            raise Exception('Cannot get return code before process is finished.')
        assert self._process is not None, "Unexpected self._process is None even though it is finished."
        return self._process.returncode

    def scriptPath(self) -> Optional[str]:
        return self._script_path

    def _remove_initial_blank_lines(self, lines: List[str]) -> List[str]:
        ii = 0
        while ii < len(lines) and len(lines[ii].strip()) == 0:
            ii = ii + 1
        return lines[ii:]

    def _get_num_initial_spaces(self, line: str) -> int:
        ii = 0
        while ii < len(line) and line[ii] == ' ':
            ii = ii + 1
        return ii
    
    @staticmethod
    def test():
        _test_shellscript()

def stop_all_scripts():
    x = list(_running_scripts.values())
    for s in x:
        s.stop()

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))

def _rmdir_with_retries(dirname, num_retries, delay_between_tries=1):
    for retry_num in range(1, num_retries + 1):
        if not os.path.exists(dirname):
            return
        try:
            shutil.rmtree(dirname)
            break
        except: # pragma: no cover
            if retry_num < num_retries:
                print('Retrying to remove directory: {}'.format(dirname))
                time.sleep(delay_between_tries)
            else:
                raise Exception('Unable to remove directory after {} tries: {}'.format(num_retries, dirname))

# def _test_error_handling_1():
#     import pytest
#     with pytest.raises(Exception):
#         ShellScript("""
#          #/bin/bash
#         echo "bad indent"
#         """)
    
#     ss = ShellScript("""
#     #!/bin/bash

#     sleep 15
#     """)
#     assert ss.elapsedTimeSinceStart() is None
#     assert ss.isRunning() == False
#     assert ss.isFinished() == False
#     ss.start()
#     assert ss.isRunning() == True
#     assert ss.isFinished() == False
#     with pytest.raises(Exception):
#         ## cannot get return code while running
#         ss.returnCode()
#     ss.stop()
#     assert ss.elapsedTimeSinceStart() < 10

#     ss.start()
#     ss.kill()
#     assert ss.elapsedTimeSinceStart() < 10

#     # it's okay to stop it if it isn't running
#     assert ss.stop() is None
#     assert ss.kill() is None
#     assert ss.stopWithSignal(signal.SIGINT, timeout=1) == True

#     for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGKILL]:
#         ss.start()
#         assert ss.isRunning() == True
#         ss.stopWithSignal(sig, timeout=0.1)
#         assert ss.isRunning() == False
#         assert ss.elapsedTimeSinceStart() < 10

def _test_coverage():
    from ._temporarydirectory import TemporaryDirectory
    with TemporaryDirectory() as tmpdir:
        _rmdir_with_retries(dirname=tmpdir + '/does-not-exist', num_retries=1)

def _test_shellscript():
    from ._temporarydirectory import TemporaryDirectory
    with TemporaryDirectory() as tmpdir:
        fname = tmpdir + '/file1.txt'
        text0 = 'some-test-text'
        ss = ShellScript(f"""
        #!/bin/bash

        echo "{text0}" > {fname}
        """)
        ss.start()
        ss.wait(timeout=5)
        with open(fname, 'r') as f:
            txt = str(f.read())
            print(f'txt = {txt}')
            assert txt == text0 + '\n'
        assert ss.returnCode() == 0
        print(f'Script path: {ss.scriptPath()}')
    
    # _test_error_handling_1()
    _test_coverage()

atexit.register(stop_all_scripts)


from typing import Any, List, Protocol
from .._daemon_connection import _daemon_url
from .._misc import _http_post_json

class RegisteredTaskFunctionCallback(Protocol):
    def __call__(self, *, channel: str, task_id: str, task_kwargs: dict, task_function_type: str):
        pass


class RegisteredTaskFunction:
    def __init__(self, *, task_function_id: str, task_function_type: str, channel: str, callback: RegisteredTaskFunctionCallback) -> None:
        self._task_function_id = task_function_id
        self._task_function_type = task_function_type
        self._channel = channel
        self._callback = callback
    @property
    def task_function_id(self):
        return self._task_function_id
    @property
    def task_function_type(self):
        return self._task_function_type
    @property
    def channel(self):
        return self._channel
    @property
    def callback(self):
        return self._callback

def register_task_functions(registered_task_functions: List[RegisteredTaskFunction], *, timeout_sec: float):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/task/registerTaskFunctions'
    # export type RegisteredTaskFunction = {
    #     channelName: string
    #     taskFunctionId: TaskFunctionId
    #     taskFunctionType: TaskFunctionType
    # }
    # export interface TaskRegisterTaskFunctionsRequest {
    #     taskFunctions: RegisteredTaskFunction[]
    #     timeoutMsec: DurationMsec
    # }
    x = []
    for a in registered_task_functions:
        x.append({
            'channelName': a.channel,
            'taskFunctionId': a.task_function_id,
            'taskFunctionType': a.task_function_type
        })
    req_data = {
        'taskFunctions': x,
        'timeoutMsec': timeout_sec * 1000
    }
    x = _http_post_json(url, req_data, headers=headers)
    # export type RequestedTask = {
    #     channelName: ChannelName
    #     taskId: TaskId
    #     taskFunctionId: TaskFunctionId
    #     kwargs: TaskKwargs
    #     taskFunctionType: TaskFunctionType
    # }
    # export interface TaskRegisterTaskFunctionsResponse {
    #     requestedTasks: RequestedTask[]
    #     success: boolean
    # }
    if not x['success']:
        raise Exception(f'Unable to register task functions. Perhaps kachery daemon is not running.')
    requested_tasks = x['requestedTasks']
    for rt in requested_tasks:
        rt_channel_name = rt['channelName']
        rt_task_id = rt['taskId']
        rt_task_function_id = rt['taskFunctionId']
        rt_task_function_type = rt['taskFunctionType']
        rt_task_kwargs = rt['kwargs']
        for a in registered_task_functions:
            if a.task_function_id == rt_task_function_id and a.channel == rt_channel_name:
                if a.task_function_type != rt_task_function_type:
                    raise Exception(f'Mismatch in task function type for {rt_task_function_id}: {a.task_function_type} <> {rt_task_function_type}')
                a.callback(channel=rt_channel_name, task_id=rt_task_id, task_function_type=rt_task_function_type, task_kwargs=rt_task_kwargs)


import json
import atexit
from kachery_client.task_backend.TaskJobManager import TaskJobManager
import multiprocessing
import random
from typing import Dict, List, Protocol, cast

from .RegisteredTaskFunction import RegisteredTaskFunction
from .RequestedTask import RequestedTask

from ._run_task_backend_worker import _run_task_backend_worker


class OnRequestedTaskCallback(Protocol):
    def __call__(self, requested_task: RequestedTask):
        pass

class TaskBackend:
    def __init__(self, registered_task_functions: List[RegisteredTaskFunction]):
        self._task_backend_id = _random_string(10)

        self._task_job_manager = TaskJobManager()
        
        run_task_backend_pipe_to_parent, run_task_backend_pipe_to_child = multiprocessing.Pipe()
        self._run_task_backend_worker_process =  multiprocessing.Process(target=_run_task_backend_worker, args=(run_task_backend_pipe_to_parent, registered_task_functions))
        self._run_task_backend_pipe_to_worker = run_task_backend_pipe_to_child

        self._on_requested_task_callbacks: List[OnRequestedTaskCallback] = []
    def start(self):
        _running_task_backends[self._task_backend_id] = self
        self._run_task_backend_worker_process.start()
    def stop(self):
        self._run_task_backend_pipe_to_worker.send({'type': 'exit'})
        self._run_task_backend_worker_process.join()

        if self._task_backend_id in _running_task_backends:
            del _running_task_backends[self._task_backend_id]
    def process_events(self):
        if self._run_task_backend_pipe_to_worker.poll():
            msg = self._run_task_backend_pipe_to_worker.recv()
            type0 = msg['type']
            if type0 == 'request_task':
                requested_task = cast(RequestedTask, msg['requested_task'])
                self._handle_requested_task(requested_task)
            else:
                raise Exception(f'Unexpected message type in task backend: {type0}')
        self._task_job_manager.process_events()
    def _handle_requested_task(self, requested_task: RequestedTask):
        task_job_manager = self._task_job_manager
        existing_job_for_task = task_job_manager.get_existing_job_for_task(requested_task)
        function_id = requested_task.registered_task_function.task_function_id
        function_type = requested_task.registered_task_function.task_function_type
        print(f'Task requested: {function_id} ({function_type})')
        if existing_job_for_task is not None:
            return
        try:
            task_output = requested_task.run()
        except Exception as e:
            error_message = f'Error calling task: {str(e)}'
            print(f'Error in {function_id}: {error_message}')
            requested_task.update_status(status='error', error_message=error_message)
            return
        if hasattr(task_output, 'status'):
            # we assume the output is a hither function
            # only import hither if we are using hither functions
            import hither2 as hi
            if isinstance(task_output, hi.Job):
                task_job_manager.add_task_job(requested_task=requested_task, job=task_output)
            else:
                raise Exception('Task output is not a hither job.')
        else:
            try:
                json.dumps(task_output)
            except:
                error_message = f'Error json-serializing result'
                print(f'Error in {function_id}: {error_message}')
                requested_task.update_status(status='error', error_message=error_message)
                return
            print(f'Finished task: {function_id}')
            requested_task.update_status(status='finished', result=task_output)

_running_task_backends: Dict[str, TaskBackend] = {}

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))

def _stop_all_task_backends():
    x = list(_running_task_backends.values())
    for s in x:
        s.stop()


atexit.register(_stop_all_task_backends)



import os
import sys

import click

import kachery_client as kc


@click.group(help="Kachery peer-to-peer command-line client")
def cli():
    pass

@click.command(help="Load or download a file.")
@click.argument('uri')
@click.option('--dest', default=None, help='Optional local path of destination file.')
def load_file(uri, dest):
    x = kc.load_file(uri, dest=dest)
    print(x)

@click.command(help="Store a file on the local node.")
@click.argument('path')
def store_file(path: str):
    x = kc.store_file(path)
    print(x)

@click.command(help="Store a link to file locally.")
@click.argument('path')
def link_file(path: str):
    x = kc.link_file(path)
    print(x)

@click.command(help="Download a file and write the content to stdout.")
@click.argument('uri')
@click.option('--start', help='The start byte (optional)', default=None)
@click.option('--end', help='The end byte non-inclusive (optional)', default=None)
def cat_file(uri, start, end):
    old_stdout = sys.stdout
    sys.stdout = None

    if start is None and end is None:
        path1 = kc.load_file(uri)
        if not path1:
            raise Exception('Error loading file for cat.')
        sys.stdout = old_stdout
        with open(path1, 'rb') as f:
            while True:
                data = os.read(f.fileno(), 4096)
                if len(data) == 0:
                    break
                os.write(sys.stdout.fileno(), data)
    else:
        assert start is not None and end is not None
        start = int(start)
        end = int(end)
        assert start <= end
        if start == end:
            return
        sys.stdout = old_stdout
        kc.load_bytes(uri=uri, start=start, end=end, write_to_stdout=True)

@click.command(help="Display kachery_client version and exit.")
def version():
    click.echo(f"This is kachery_client version {kc.__version__}")
    exit()

cli.add_command(cat_file)
cli.add_command(load_file)
cli.add_command(store_file)
cli.add_command(link_file)
cli.add_command(version)



import time
from multiprocessing.connection import Connection
from typing import List, Protocol

from .RegisteredTaskFunction import RegisteredTaskFunction
from .RequestedTask import RequestedTask

from .._daemon_connection import _daemon_url
from .._misc import _http_post_json


def _run_task_backend_worker(pipe_to_parent: Connection, registered_task_functions: List[RegisteredTaskFunction]):
    while True:
        while pipe_to_parent.poll():
            x = pipe_to_parent.recv()
            if isinstance(x, dict):
                type0 = x.get('type', '')
                if type0 == 'exit':
                    return
                else:
                    raise Exception(f'Unexpected message type in _run_task_backend_worker: {type0}')
            else:
                print(x)
                raise Exception('Unexpected message in _run_task_backend_worker')
        requested_tasks = _register_task_functions(registered_task_functions, timeout_sec=4)
        for requested_task in requested_tasks:
            pipe_to_parent.send({
                'type': 'request_task',
                'requested_task': requested_task
            })
        time.sleep(0.1)

def _register_task_functions(registered_task_functions: List[RegisteredTaskFunction], *, timeout_sec: float):
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/task/registerTaskFunctions'
    # export type RegisteredTaskFunction = {
    #     channelName: string
    #     taskFunctionId: TaskFunctionId
    #     taskFunctionType: TaskFunctionType
    # }
    # export interface TaskRegisterTaskFunctionsRequest {
    #     taskFunctions: RegisteredTaskFunction[]
    #     timeoutMsec: DurationMsec
    # }
    x = []
    for a in registered_task_functions:
        x.append({
            'channelName': a.channel,
            'taskFunctionId': a.task_function_id,
            'taskFunctionType': a.task_function_type
        })
    req_data = {
        'taskFunctions': x,
        'timeoutMsec': timeout_sec * 1000
    }
    x = _http_post_json(url, req_data, headers=headers)
    # export type RequestedTask = {
    #     channelName: ChannelName
    #     taskId: TaskId
    #     taskFunctionId: TaskFunctionId
    #     kwargs: TaskKwargs
    #     taskFunctionType: TaskFunctionType
    # }
    # export interface TaskRegisterTaskFunctionsResponse {
    #     requestedTasks: RequestedTask[]
    #     success: boolean
    # }
    if not x['success']:
        raise Exception(f'Unable to register task functions. Perhaps kachery daemon is not running.')
    requested_tasks = x['requestedTasks']
    ret: List[RequestedTask] = []
    for rt in requested_tasks:
        rt_channel_name = rt['channelName']
        rt_task_id = rt['taskId']
        rt_task_function_id = rt['taskFunctionId']
        rt_task_function_type = rt['taskFunctionType']
        rt_task_kwargs = rt['kwargs']
        for x in registered_task_functions:
            if x.channel == rt_channel_name and x.task_function_id == rt_task_function_id:
                if x.task_function_type == rt_task_function_type:
                    ret.append(RequestedTask(
                        registered_task_function=x,
                        kwargs=rt_task_kwargs,
                        task_id=rt_task_id
                    ))
                else:
                    print(f'Warning: mismatch in task function type for {rt_task_function_id}: {x.task_function_type} <> {rt_task_function_type}')
    return ret



from .taskfunction import find_taskfunction

class RegisteredTaskFunction:
    def __init__(self, *, task_function_id: str, task_function_type: str, channel: str) -> None:
        self._task_function_id = task_function_id
        self._task_function_type = task_function_type
        self._channel = channel
    @property
    def task_function_id(self):
        return self._task_function_id
    @property
    def task_function_type(self):
        return self._task_function_type
    @property
    def channel(self):
        return self._channel
    def run(self, *, kwargs: dict):
        f = find_taskfunction(self.task_function_id)
        if f is None:
            raise Exception(f'Unable to find task function: {self.task_function_id}')
        return f(**kwargs)


import setuptools

setuptools.setup(
    packages=setuptools.find_packages(),
    include_package_data=True,
    scripts=[
        "bin/kachery-client",
        "bin/kachery-cat",
        "bin/kachery-load",
        "bin/kachery-store",
        "bin/kachery-link"
    ],
    install_requires=[
        "click",
        "simplejson",
        "requests",
        "jinjaroot"
    ]
)



import os
import sys
import hashlib
import shutil
import random
import json
from typing import Optional, Tuple, Union
from ._misc import _parse_kachery_uri
from ._daemon_connection import _kachery_storage_dir


def _local_kachery_storage_load_file(*, sha1_hash: str):
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    path = _get_path_ext(hash=sha1_hash, create=False, directory=sha1_directory)
    if os.path.exists(path):
        return path
    elif os.path.exists(path + '.link'):
        linked_file_path = _find_linked_file(path + '.link')
        if linked_file_path is not None:
            return linked_file_path
    return None

def _find_linked_file(link_path: str):
    with open(link_path, 'r') as f:
        link: dict = json.load(f)
    path = link.get('path', None)
    if path is None:
        print(f'WARNING: problem with link (no path field): {link_path}')
        return None
    stat = link.get('stat', None)
    if stat is None:
        print(f'WARNING: problem with link (no stat field): {link_path}')
        return None
    if os.path.exists(path):
        if _stat_matches(stat, path):
            return path
        else:
            print(f'WARNING: linked file may have been modified: {link_path} {path}')
            return None
    else:
        print('WARNING: linked file does not exist: {link_path} {path}')
    return None

def _stat_matches(stat: dict, path: str):
    s = os.stat(path)
    if stat['size'] != s.st_size:
        return False
    if stat['mtime'] != s.st_mtime:
        return False
    return True
    
def _local_kachery_storage_load_bytes(*, sha1_hash: str, start: Union[int, None]=None, end: Union[int, None]=None, write_to_stdout: bool=False):
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    path = _get_path_ext(hash=sha1_hash, create=False, directory=sha1_directory)
    if os.path.exists(path):
        return _load_bytes_from_local_file(local_fname=path, start=start, end=end, write_to_stdout=write_to_stdout)
    else:
        return None

def _local_kachery_storage_store_file(*, path: str, _no_manifest=False) -> Tuple[str, str, Union[str, None]]:
    from ._store_file import _store_json # don't want circular dependencies
    if (not _no_manifest) and (os.path.getsize(path) > 20000000):
        hash0, manifest0 = _compute_local_file_sha1_and_manifest(path)
        if manifest0 is None:
            raise Exception(f'Unable to compute hash of file: {path}')
        manifest_uri = _store_json(manifest0)
        protocol, algorithm, manifest_hash, additional_path, query = _parse_kachery_uri(manifest_uri)
    else:
        hash0 = _get_file_hash(path)
        manifest_hash = None
    assert hash0 is not None
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    path0 = _get_path_ext(hash=hash0, create=True, directory=sha1_directory)
    if not os.path.exists(path0):
        tmp_path = path0 + '.copying.' + _random_string(6)
        shutil.copyfile(path, tmp_path)
        _rename_file(tmp_path, path0, remove_if_exists=False)
    return path0, hash0, manifest_hash

def _local_kachery_storage_link_file(*, path: str, _no_manifest=False) -> Tuple[str, str, Union[str, None]]:
    from ._store_file import _store_json # don't want circular dependencies
    if (not _no_manifest) and (os.path.getsize(path) > 20000000):
        hash0, manifest0 = _compute_local_file_sha1_and_manifest(path)
        if manifest0 is None:
            raise Exception(f'Unable to compute hash of file: {path}')
        manifest_uri = _store_json(manifest0)
        protocol, algorithm, manifest_hash, additional_path, query = _parse_kachery_uri(manifest_uri)
    else:
        hash0 = _get_file_hash(path)
        manifest_hash = None
    assert hash0 is not None
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    path0 = _get_path_ext(hash=hash0, create=True, directory=sha1_directory)
    if not os.path.exists(path0):
        tmp_path = path0 + '.link.' + _random_string(6)
        with open(tmp_path, 'w') as f:
            json.dump({
                'path': os.path.abspath(path),
                'manifest_hash': manifest_hash,
                'stat': {
                    'size': os.stat(path).st_size,
                    'mtime': os.stat(path).st_mtime
                }
            }, f)
        _rename_file(tmp_path, path0 + '.link', remove_if_exists=True)
    return path0, hash0, manifest_hash

def _get_file_hash(path: str, *, _cache_only=False):
    algorithm = 'sha1'
    if os.path.getsize(path) < 100000:
        # if it is a small file, we can compute the hash directly
        # this is important when the kachery storage dir is on a remote file system
        return _compute_file_hash(path, algorithm=algorithm)
    path = os.path.abspath(path)
    basename = os.path.basename(path)
    sha1_directory = f'{_kachery_storage_dir()}/sha1'
    if len(basename) == _length_of_hash_for_algorithm(algorithm):
        # suspect it is itself a file in the cache
        if _get_path_ext(hash=basename, create=False, directory=sha1_directory) == path:
            # in that case we don't need to compute
            return basename

    if _cache_only:
        return None
    hash1 = _compute_file_hash(path, algorithm=algorithm)
    
    if not hash1:
        return None

    return hash1

def _compute_file_hash(path: str, algorithm: str) -> Optional[str]:
    if not os.path.exists(path):
        return None
    if (os.path.getsize(path) > 1024 * 1024 * 100):
        print('Computing {} of {}'.format(algorithm, path))
    BLOCKSIZE = 65536
    hashsum = getattr(hashlib, algorithm)()
    with open(path, 'rb') as file:
        buf = file.read(BLOCKSIZE)
        while len(buf) > 0:
            hashsum.update(buf)
            buf = file.read(BLOCKSIZE)
    return hashsum.hexdigest()

def _length_of_hash_for_algorithm(algorithm):
    if algorithm == 'sha1':
        return 40
    elif algorithm == 'md5':
        return 32
    else:
        raise Exception('Unexpected algorithm: {}'.format(algorithm))

def _get_path_ext(hash: str, *, create: bool=True, directory: str) -> str:
    path1 = os.path.join(hash[0:2], hash[2:4], hash[4:6])
    path0 = os.path.join(str(directory), path1)
    if create:
        if not os.path.exists(path0):
            try:
                os.makedirs(path0)
            except:
                if not os.path.exists(path0):
                    raise Exception('Unable to make directory: ' + path0)
    return os.path.join(path0, hash)

def _load_bytes_from_local_file(local_fname: str, *, start: Union[int, None]=None, end: Union[int, None]=None, write_to_stdout: bool=False) -> Union[bytes, None]:
    size0 = os.path.getsize(local_fname)
    if start is None:
        start = 0
    if end is None:
        end = size0
    if start < 0 or start > size0 or end < start or end > size0:
        raise Exception('Invalid start/end range for file of size {}: {} - {}'.format(size0, start, end))
    if start == end:
        return bytes()
    with open(local_fname, 'rb') as f:
        f.seek(start)
        if write_to_stdout:
            ii = start
            while ii < end:
                nn = min(end - ii, 4096)
                data0 = f.read(nn)
                ii = ii + nn
                sys.stdout.buffer.write(data0)
            return None
        else:
            return f.read(end-start)

def _rename_file(path1: str, path2: str, remove_if_exists: bool) -> None:
    if os.path.abspath(path1) == os.path.abspath(path2):
        return
    if os.path.exists(path2):
        if remove_if_exists:
            try:
                os.unlink(path2)
            except:
                # maybe it was removed by someone else
                pass
        else:
            # already exists, let's just let it be
            return
    try:
        os.rename(path1, path2)
    except:
        if os.path.exists(path2):
            if not remove_if_exists:
                # all good
                return
            raise Exception('Problem renaming file: {} -> {}'.format(path1, path2))
        else:
            raise Exception('Problem renaming file:: {} -> {}'.format(path1, path2))

def _compute_local_file_sha1_and_manifest(path):
    algorithm = 'sha1'
    manifest = {
        'size': 0,
        'sha1': '',
        'chunks': []
    }
    if not os.path.exists(path):
        return None, None
    size0 = os.path.getsize(path)
    if (size0 > 1024 * 1024 * 100):
        print('Computing {} and manifest of {}'.format(algorithm, path))

    chunk_size = 20000000

    hashsum = getattr(hashlib, algorithm)()
    with open(path, 'rb') as file:
        pos = 0
        while pos < size0:
            
            this_chunk_size = min(chunk_size, size0 - pos)

            this_chunk_hashsum = getattr(hashlib, algorithm)()
            buf = file.read(this_chunk_size)
            this_chunk_hashsum.update(buf)
            
            hashsum.update(buf)
            
            chunk = {
                'start': pos,
                'end': pos + this_chunk_size,
                'sha1': this_chunk_hashsum.hexdigest()
            }
            manifest['chunks'].append(chunk)
            
            pos = pos + this_chunk_size
            
    sha1 = hashsum.hexdigest()
    manifest['sha1'] = sha1
    manifest['size'] = size0
    return sha1, manifest

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


from typing import Any, Union
from .RegisteredTaskFunction import RegisteredTaskFunction
from ._update_task_status import _update_task_status


class RequestedTask:
    def __init__(self, *, registered_task_function: RegisteredTaskFunction, kwargs: dict, task_id: str) -> None:
        self._registered_task_function = registered_task_function
        self._kwargs = kwargs
        self._task_id = task_id
        self._status = 'waiting'
    @property
    def registered_task_function(self):
        return self._registered_task_function
    @property
    def kwargs(self):
        return self._kwargs
    @property
    def task_id(self):
        return self._task_id
    @property
    def task_function_type(self):
        return self._registered_task_function.task_function_type
    def run(self):
        return self._registered_task_function.run(kwargs=self._kwargs)
    @property
    def status(self):
        return self._status
    def update_status(self, *, status: str, error_message: Union[str, None]=None, result: Union[Any, None]=None):
        self._status = status
        _update_task_status(channel=self.registered_task_function.channel, task_id=self.task_id, task_function_type=self.task_function_type, status=status, result=result, error_message=error_message)


from typing import Any, Tuple, Union
from ._daemon_connection import _daemon_url
from ._misc import _http_post_json, _http_get_json

class OutgoingTaskRequest:
    def __init__(self, *, channel: str, task_id: str, task_function_type: str, task_result_url: Union[str, None], status: str, error_message: Union[str, None], query_result: Union[Any, None]):
        self._channel = channel
        self._task_id = task_id
        self._task_function_type = task_function_type
        self._task_result_url = task_result_url
        self._status = status
        self._error_message = error_message
        self._query_result = query_result
        self._downloaded_result: Union[Any, None] = None
    @property
    def status(self):
        return self._status
    @property
    def task_result_url(self):
        return self._task_result_url
    @property
    def task_result(self):
        if self._status != 'finished':
            raise Exception('Cannot get task result if status is not finished')
        if self._downloaded_result:
            return self._downloaded_result
        if not self._task_result_url:
            raise Exception('No task result url')
        self._downloaded_result = _http_get_json(self._task_result_url)
        return self._downloaded_result
    @property
    def query_result(self):
        if self._status != 'finished':
            raise Exception('Cannot get query result if status is not finished')
        return self._query_result
    @property
    def error_message(self):
        return self._error_message
    def wait(self, timeout_sec: float):
        if self._status not in ['finished', 'error']:
            daemon_url, headers = _daemon_url()
            # export type TaskWaitForTaskResultRequest = {
            #     channelName: ChannelName
            #     taskId: TaskId
            #     taskFunctionType: TaskFunctionType,
            #     timeoutMsec: DurationMsec
            # }
            url = f'{daemon_url}/task/waitForTaskResult'
            req_data = {
                'channelName': self._channel,
                'taskId': self._task_id,
                'taskFunctionType': self._task_function_type,
                'timeoutMsec': timeout_sec * 1000
            }
            x = _http_post_json(url, req_data, headers=headers)
            if not x['success']:
                print(x)
                raise Exception(f'Unable to wait on task')
            # export type TaskWaitForTaskResultResponse = {
            #     success: boolean
            #     status: TaskStatus
            #     errorMessage?: ErrorMessage
            #     taskResultUrl?: UrlString
            #     queryResult?: JSONValue
            # }
            self._status = x['status']
            self._task_result_url = x.get('taskResultUrl', None)
            self._error_message = x.get('errorMessage', None)
            self._query_result = x.get('queryResult', None)
        if self._status == 'error':
            raise Exception(f'Task error: {self._error_message}')
        if self._status == 'finished':
            if self._task_function_type == 'pure-calculation':
                return self.task_result
            elif self._task_function_type == 'query':
                return self.query_result
            elif self._task_function_type == 'action':
                return True
            else:
                raise Exception(f'Unexpected function type: {self._task_function_type}')
        return None

def request_task(*, task_function_id: str, task_kwargs: dict, task_function_type: str, channel: str) -> OutgoingTaskRequest:
    daemon_url, headers = _daemon_url()
    url = f'{daemon_url}/task/requestTask'
    # export type TaskRequestTaskRequest = {
    #     channelName: ChannelName
    #     taskFunctionId: TaskFunctionId
    #     taskKwargs: TaskKwargs
    #     taskFunctionType: TaskFunctionType
    #     timeoutMsec: DurationMsec
    # }
    req_data = {
        'channelName': channel,
        'taskFunctionId': task_function_id,
        'taskKwargs': task_kwargs,
        'taskFunctionType': task_function_type,
        'timeoutMsec': 1000
    }
    x = _http_post_json(url, req_data, headers=headers)
    # export type TaskRequestTaskResponse = {
    #     success: boolean
    #     taskId: TaskId,
    #     status: TaskStatus
    #     errorMessage?: ErrorMessage
    #     taskResultUrl?: UrlString
    #     queryResult?: JSONValue
    # }
    if not x['success']:
        raise Exception(f'Unable to load task result')
    status = x['status']
    task_id = x['taskId']
    task_result_url = x.get('taskResultUrl', None)
    query_result = x.get('queryResult', None)
    error_message = x.get('errorMessage', None)
    return OutgoingTaskRequest(channel=channel, task_id=task_id, task_function_type=task_function_type, task_result_url=task_result_url, status=status, error_message=error_message, query_result=query_result)